﻿using System;
using System.Net;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Documents;
using System.Windows.Ink;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Animation;
using System.Windows.Shapes;
using System.Threading;

namespace DetectCPUCoreCount
{
    public class PerformanceMeasure
    {
        public class ThreadData
        {
            public long CountOfOperations = 0;
            public Thread Thread;
            public object LockObject;
        }

        private static int _threadsRunning = 0;
        private static bool _start = false;

        /// <summary>
        /// Determines average operation count executed for a specific time period for a number of threads
        /// Use to scale core detection algorithm based on machine speed
        /// </summary>
        /// <param name="threadCount">threads to run</param>
        /// <param name="milliseconds">amount of time to run each thread</param>
        /// <returns>average operations per thread</returns>
        public static long GetAverageOperations(int threadCount, int milliseconds)
        {
            ThreadData[] data = new ThreadData[threadCount];

            _start = false;
            _threadsRunning = threadCount;
            DateTime startTime = DateTime.Now;
            for (int i = 0; i < threadCount; i++)
            {
                data[i] = new ThreadData();
                data[i].LockObject = data;
                data[i].Thread = new Thread(OperationCountThread);
                data[i].Thread.Start(data[i]);
            }

            _start = true;
            while ((DateTime.Now - startTime).TotalMilliseconds < milliseconds) Thread.Sleep(1);
            _start = false;
            Thread.Sleep(20);
            long totalCount = 0;
            for (int i = 0; i < threadCount; i++)
            {
                totalCount += data[i].CountOfOperations;
            }

            // calculate how far off we were in the waits
            double factor = (DateTime.Now - startTime).TotalMilliseconds / milliseconds;

            return (long) ( (totalCount / threadCount) / factor );
        }

        /// <summary>
        /// Gets the amount of operations that can be executed over a specific time period
        /// </summary>
        /// <param name="milliseconds">time to execute the operations</param>
        /// <param name="threadCount">number of threads to use</param>
        /// <returns>milliseconds (absolute time, not thread time) used to compute countToTryWith*threadCount operations</returns>
        public static double GetAbsoluteTimeToCompute(int threadCount, long countToTryWith)
        {
            ThreadData[] data = new ThreadData[threadCount];

            _start = false;
            _threadsRunning = threadCount;
            DateTime startTime = DateTime.Now;
            for (int i = 0; i < threadCount; i++)
            {
                data[i] = new ThreadData();
                data[i].LockObject = data;
                data[i].CountOfOperations = countToTryWith;
                data[i].Thread = new Thread(MeasureThread);
                if (i > 0)
                {
                    data[i].Thread.Start(data[i]);
                }
            }

            _start = true;
            MeasureThread(data[0]);

            while (true)
            {
                lock (data) {
                    if (_threadsRunning <= 0) break;
                }
                Thread.Sleep(10);
            }
            double totalAbsoluteMsec = (DateTime.Now - startTime).TotalMilliseconds;

            double totalComputations = countToTryWith * threadCount;
            return totalAbsoluteMsec; // the amount of computations executed in this excercise :)
        }

        private static void MeasureThread(object obj)
        {
            ThreadData data = (ThreadData)obj;
            while (!_start) { };
            while (data.CountOfOperations > 0)
            {
                data.CountOfOperations--;
            }
            lock (data.LockObject)
            {
                _threadsRunning--;
            }
        }

        private static void OperationCountThread(object obj)
        {
            ThreadData data = (ThreadData)obj;
            while (!_start) { Thread.Sleep(1); }
            while (_start)
            {
                data.CountOfOperations++;
            }
        }

        /// <summary>
        /// base cycle length in milliseconds
        /// </summary>
        public const int MSEC_PER_CYCLE = 150;

        /// <summary>
        /// acceptable threshold for using a measurement
        /// </summary>
        public const int MSEC_ACCEPTABLE_THRESHOLD = 100;

        /// <summary>
        /// times that the operation shoul slowdown in order to stop the algorithm and retrieve the number of cores
        /// </summary>
        public const double SLOWDOWN_THRESHOLD_MULTIPLIER = 1.9;

        /// <summary>
        /// Returns the number of cores (optimal number of threads to run simultaneously) on the system
        /// </summary>
        /// <returns></returns>
        public static int GetCoreCount()
        {
            long COUNT_OF_OPERATIONS;
            double msecElapsed;
            COUNT_OF_OPERATIONS = PerformanceMeasure.GetAverageOperations(1, MSEC_PER_CYCLE);
            // get correct count of operations: work around for slowdowns on startup or other (lightweight) apps running in the background
            while (PerformanceMeasure.GetAbsoluteTimeToCompute(1, COUNT_OF_OPERATIONS) < MSEC_ACCEPTABLE_THRESHOLD) 
            {
                COUNT_OF_OPERATIONS = PerformanceMeasure.GetAverageOperations(1, MSEC_PER_CYCLE);
            }
            
            int[] coresToDetect = { 1, 2, 4, 8, 16, 32 };
            bool first = true;
            double prevValue = 0;
            int coreCount = 2;
            int prevI = 2;
            foreach (int i in coresToDetect)
            {
                msecElapsed = PerformanceMeasure.GetAbsoluteTimeToCompute(i, COUNT_OF_OPERATIONS);
                if (!first)
                {
                    if (((msecElapsed / prevValue) > SLOWDOWN_THRESHOLD_MULTIPLIER) && (msecElapsed >= MSEC_ACCEPTABLE_THRESHOLD))
                    {
                        coreCount = prevI;
                        break;
                    }
                }
                prevValue = msecElapsed;
                prevI = i;
                first = false;
            }

            return coreCount;
        }
    }
}
